/*
 * Copyright (c) 1987 Tom Anderson; 20831 Frank Waters Road;
 * Stanwood, WA  98282.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Tom Anderson not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  Tom Anderson makes
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * TOM ANDERSON DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO
 * EVENT SHALL TOM ANDERSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 */

/*
 * manage the board state
 *
 */
static char copyright[] = "Copyright 1987 Tom Anderson";

#include <stdio.h>
#include <sys/types.h>

#include <X11/Intrinsic.h>
#include "mines.h"

#define srandom srand48
#define random lrand48

extern char *malloc();

Square **MainBoard;
int BoardRows;
int BoardCols;
Square *PlayerSquare;
BOOL GameOver;
BOOL Impossible;

static int lastlevel;

static char *
s_malloc(size)
int size;
{
	char *p;

	p = malloc(size);
	if (!p) {
		fprintf(stderr,"Can't allocate memory for playing board\n");
		exit(1);
	}
	return p;
}

void
AllocBoard(numrows,numcols,deflevel)
int numrows,numcols;
int deflevel;   /* default number of mines to put in */
{
	int i;

	BoardRows = numrows;
	BoardCols = numcols;

	if (BoardRows < 3)
		BoardRows = 3;
	if (BoardCols < 3)
		BoardCols = 3;
	if (BoardRows > MAX_BOARD_ROWS)
		BoardRows = MAX_BOARD_ROWS;
	if (BoardCols > MAX_BOARD_COLS)
		BoardCols = MAX_BOARD_COLS;

	MainBoard = (Square **)s_malloc(BoardRows * sizeof(Square *));
	for (i=0; i<BoardRows; i++) {
		MainBoard[i] = (Square *)s_malloc(BoardCols * sizeof(Square));
	}
	if (deflevel>0) /* set default mine quantity */
		lastlevel = deflevel;
	else
		lastlevel = (BoardRows + BoardCols) / 2;
}

/*
 * set up the playing surface at the beginning of the game
 */
void
InitBoard(level)
int level;                      /* number of mines */
{
	register int i, j, minesPlaced, cutoff;
	register Square * sqp;
	int pass = 0;
	void MazeConnected();
	char *MineWarningMessage();

	if (level<=0) level=lastlevel;
	if (level > (BoardRows*BoardCols-5))
		level = BoardRows*BoardCols-5;  /* avoid infinite loop below */
	lastlevel = level;
	SetMineCountDisplay(level);

	/* zero the board */
	for (i = 0 ; i < BoardRows ; i++) {
		for (j = 0 ; j < BoardCols ; j++) {
			sqp = &MainBoard[i][j];
			sqp->traversed = sqp->mined = sqp->occupied
			    = sqp->unsafe = sqp->safe = FALSE;
			sqp->nearby = 0;
			sqp->visited = FALSE;
			sqp->row = i;
			sqp->col = j;
		}
	}

	/* make several passes over the board, placing mines at
	 * a probability calculated to require about 3 passes on the average
	 */
	cutoff = level / 3 + 1;
	for (minesPlaced = 0 ; minesPlaced < level ; ) {
		for (i = 0 ; i < BoardRows ; i++) {
			for (j = 0 ; j < BoardCols ; j++) {
				if ((random() % (BoardRows * BoardCols)) <cutoff
				    && ! MainBoard[i][j].mined
				    && minesPlaced < level
				    && ! (i <= 1 && j <= 1)
				    && ! (i==BoardRows-1 && j==BoardCols-1)) {
					MainBoard[i][j].mined = TRUE;
					minesPlaced++;
				}
			}
		}
	}
	PlayerSquare = &MainBoard[0][0];
	MainBoard[0][0].occupied = MainBoard[0][0].traversed = TRUE;
	GameOver = FALSE;

	MazeConnected();

	Message(MineWarningMessage());

#ifndef NOSCORE
	SetHints(0);
	SetNumberMoves(0);
	SetMinesMarked(0);
	SetSafesMarked(0);
#endif
}

BOOL
Traverse(x, y)
int x;
int y;
{
	int i;
	int j;

	if (y == BoardRows-1 && x == BoardCols-1) return TRUE;

	MainBoard[y][x].visited = TRUE;

	for (i=y-1; i <= y+1; i++) {
		if (i < 0 || i == BoardRows) continue;
		for (j=x-1; j <= x+1; j++) {
			if (j < 0 || j == BoardCols) continue;
			if (i == y && j == x) continue;
			if (MainBoard[i][j].visited) continue;
			if (MainBoard[i][j].mined) continue;

			if (Traverse(j, i)) return TRUE;
		}
	}

	return FALSE;
}

void
MazeConnected()
{
	Impossible = !Traverse(0, 0);
}

SetCellWidget(row,col,w)
int row,col;
Widget w;
{
	MainBoard[row][col].w = w;
}

/* translate (row,col) to Square pointer */
Square *
GetSquare(row,col)
int row,col;
{
	return (&MainBoard[row][col]);
}

/* translate Widget to Square pointer; NULL pointer if didn't get it */
Square *
GetWidgetSquare(w)
Widget w;
{
	int x,y;
	Square *sqp;

	for (y=0; y<BoardRows; y++) {
		for (x=0; x<BoardCols; x++) {
			sqp = &MainBoard[y][x];
			if (sqp->w == w) {
				return sqp;
			}
		}
	}
	return NULL;
}

/* toggle a square's marking as being probably safe or unsafe */
void
MarkSquare(sqp,safe)
Square *sqp;
BOOL safe;
{
	if (sqp->traversed)
		return;

	if (safe) {
#ifndef NOSCORE
		IncrSafesMarked(sqp->safe ? -1 : 1);
		IncrMinesMarked(sqp->unsafe ? -1 : 0);
#endif
		sqp->safe = ! sqp->safe;
		sqp->unsafe = FALSE;
	} else {
#ifndef NOSCORE
		IncrSafesMarked(sqp->safe ? -1 : 0);
		IncrMinesMarked(sqp->unsafe ? -1 : 1);
#endif
		sqp->unsafe = ! sqp->unsafe;
		sqp->safe = FALSE;
	}
	DrawSquare(sqp);
}


/* try to move to a certain board coordinate */
void
DoMove(sqp)
Square *sqp;
{
	int row,col;
	register int y = PlayerSquare->row;
	register int x = PlayerSquare->col;

        char *MineWarningMessage(); 
	row = sqp->row;
	col = sqp->col;
	if ((col == x) && (row == y)) return;
	if (MainBoard[row][col].unsafe) return;

	if ((abs(col - x) > 1) || (abs(row - y) > 1)) {
		int xx, yy;

		/* allow moving to any square which is or is adjacent to
		 * a traversed square */
		for (xx = col - 1 ; xx <= col + 1 ; xx++) {
			for (yy = row - 1 ; yy <= row + 1 ; yy++) {
				if (xx >= 0 && xx < BoardCols &&
				    yy >= 0 && yy < BoardRows &&
				    MainBoard[yy][xx].traversed) goto ok;
			}
		}
		return;
	}

ok:

#ifndef NOSCORE
	IncrNumberMoves(1);
#endif

	/* step off our current square */
	MainBoard[y][x].occupied = FALSE;
	DrawSquare(PlayerSquare);
	sqp = &MainBoard[row][col];

	/* if we stepped on a mine, blow him up */
	if (sqp->mined) {
		GameOver = TRUE;
		Message("You just exploded.  ");
		DrawBoard(0);
	}

	/* else if this is home, render congratulations */
	else if (col == BoardCols-1 && row == BoardRows-1) {
		PlayerSquare = sqp;
		(void)MineWarningMessage();     /* fix up nearby count */
		sqp->traversed = sqp->occupied = TRUE;
		GameOver = TRUE;
		Message("You honestly made it!");
		DrawBoard(0);
	}

	/* else move onto the new square */
	else {
		/* if player marked that square as safe, unmark */
		if (sqp->safe) {
#ifndef NOSCORE
			IncrSafesMarked(-1);
#endif
			sqp->safe = FALSE;
		}
		PlayerSquare = sqp;
		Message(MineWarningMessage());
		sqp->traversed = sqp->occupied = TRUE;
		DrawSquare(sqp);
	}
}

Square *
GetHintSquare() /* returns pointer to a square suitable for a hint; or NULL */
{
	int i;
	int j;
	int x = PlayerSquare->col;
	int y = PlayerSquare->row;

	for (j = y-1; j <= y+1; j++) {
		if (j < 0 || j == BoardRows) continue;
		for (i = x-1; i <= x+1; i++) {
			if (i < 0 || i == BoardCols) continue;
			if (j == y && i == x) continue;

			if (MainBoard[j][i].traversed) continue;

			if (!MainBoard[j][i].mined) {

#ifndef NOSCORE
				IncrHints(1);
#endif
				MarkSquare(&MainBoard[j][i],1);
					/* mark as safe, keep track of counts */

				return (GetSquare(j,i));
			}
		}
	}
	return NULL;
}

/*
 * return a pointer to the warning message
 */
char *
MineWarningMessage()
{
	static char warning[128];
	register int x, y;
	int minesFound = 0;

	if (!PlayerSquare->traversed) {
		for (x = PlayerSquare->col - 1 ;
		     x <= PlayerSquare->col + 1 ;
		     x++) {
			for (y = PlayerSquare->row - 1 ;
			     y <= PlayerSquare->row + 1 ;
			     y++) {
				if (x >= 0 && x < BoardCols &&
				    y >= 0 && y < BoardRows &&
				    MainBoard[y][x].mined)
					minesFound++;
			}
		}
		PlayerSquare->nearby = minesFound;
	} else {
		minesFound = PlayerSquare->nearby;
	}

	if (minesFound == 0) {
		sprintf(warning, "No mines nearby.");
	} else if (minesFound == 1) {
		sprintf(warning, "1 mine nearby.");
	} else {
		sprintf(warning, "%d mines nearby.", minesFound);
	}

	return(warning);
}

/* end */
